class NodesController < InheritedResources::Base
  belongs_to :node_class, :optional => true
  belongs_to :node_group, :optional => true
  respond_to :html, :yaml, :json
  before_filter :raise_if_enable_read_only_mode, :only => [:new, :edit, :create, :update, :destroy]

  layout lambda {|c| c.request.xhr? ? false : 'application' }

  include ConflictAnalyzer
  include ConflictHtml
  include ClassParametersHelper

  def index
    raise NodeClassificationDisabledError.new if !SETTINGS.use_external_node_classification and request.format == :yaml
    scoped_index :unhidden
  end

  [:unreported, :failed, :unresponsive, :pending, :changed, :unchanged].each do |action|
    define_method(action) {scoped_index :unhidden, action}
  end

  def new
    new! do |format|
      format.html {
        set_group_and_class_autocomplete_data_sources(@node)
        @class_parameters = {}
      }
    end
  end

  def create
    ActiveRecord::Base.transaction do

      create! do |success, failure|
        success.html {
          node = Node.find_by_name(params[:node][:name])
          store_class_parameters(node, node.node_class_memberships, params[:class_parameters])

          unless(force_create?)

            new_conflicts_message = get_new_conflicts_message_as_html({}, node)
            if new_conflicts_message
              render_conflicts_html new_conflicts_message, "Create", "jQuery('#force_create').attr('value', 'true'); jQuery('#submit_button').click();"
            end
          end

          render :json => { :status => "ok", :valid => "true", :redirect_to => url_for(node) }, :content_type => 'application/json'
        };

        failure.html {
          set_group_and_class_autocomplete_data_sources(@node)
          html = render_to_string(:template => "shared/_error",
                                  :layout => false,
                                  :locals => { :object => @node })
          render :json => { :status => "error", :error_html => html }, :content_type => 'application/json'
        }
      end
    end
  end

  def hidden
    scoped_index :hidden
  end

  def search
    index! do |format|
      format.html {
        @search_params = params['search_params'] || []
        @search_params.delete_if {|param| param.values.any?(&:blank?)}

        inventory_error = nil
        nodes = []

        unless @search_params.empty?
          begin
            nodes = Node.find_from_inventory_search(@search_params)
          rescue => e
            inventory_error = e
          end
        end

        set_collection_ivar(nodes)
        render :inventory_search, :locals => {:inventory_error => inventory_error}
      }
    end
  end

  def show
    begin
      raise NodeClassificationDisabledError.new if !SETTINGS.use_external_node_classification and request.format == :yaml
      show!
    rescue ActiveRecord::RecordNotFound => e
      raise e unless request.format == :yaml
      render :yaml => {'classes' => []}
    rescue ParameterConflictError => e
      raise e unless request.format == :yaml
      render :text => "Node \"#{resource.name}\" has conflicting variable(s): #{resource.errors[:parameters].to_a.to_sentence}", :content_type => 'text/plain', :status => 500
    rescue ClassParameterConflictError => e
      raise e unless request.format == :yaml
      render :text => "Node \"#{resource.name}\" has conflicting class parameter(s): #{resource.errors[:classParameters].to_a.to_sentence}", :content_type => 'text/plain', :status => 500
    rescue NodeClassificationDisabledError => e
      render :text => "Node classification has been disabled", :content_type => 'text/plain', :status => 403
    end
  end

  def edit
    edit! do |format|
      format.html {
        set_group_and_class_autocomplete_data_sources(@node)
        @class_parameters = get_class_parameters(@node.node_class_memberships)
      }
    end
  end

  def update
    ActiveRecord::Base.transaction do
      # Before performing the update, we want to get the existing conflicts so
      # we can later determine if we've introduced new conflicts.
      old_conflicts = force_update? ? nil : get_current_conflicts(resource)

      # Unset the instance variable after we've used it so that update! will
      # reload it fresh
      set_resource_ivar(nil)

      update! do |success, failure|
        success.html {
          store_class_parameters(resource, resource.node_class_memberships, params[:class_parameters])

          unless(force_update?)

            new_conflicts_message = get_new_conflicts_message_as_html(old_conflicts, resource)
            if new_conflicts_message
              render_conflicts_html new_conflicts_message, "Update", "jQuery('#force_update').attr('value', 'true'); jQuery('#submit_button').click();"
            end
          end

          render :json => { :status => "ok", :valid => "true", :redirect_to => url_for(resource) }, :content_type => 'application/json'
        };

        failure.html {
          set_group_and_class_autocomplete_data_sources(resource)
          html = render_to_string(:template => "shared/_error",
                                  :layout => false,
                                  :locals => { :object => resource })
          render :json => { :status => "error", :error_html => html }, :content_type => 'application/json'
        }
      end
    end
  end

  def hide
    respond_to do |format|
      resource.hidden = true
      resource.save!

      format.html { redirect_to node_path(resource) }
    end
  end

  def unhide
    respond_to do |format|
      resource.hidden = false
      resource.save!

      format.html { redirect_to node_path(resource) }
    end
  end

  def facts
    respond_to do |format|
      format.html {
        begin
          render :partial => 'nodes/facts', :locals => {:node => resource, :facts => resource.facts}
        rescue => e
          render :text => "Could not retrieve facts from inventory service: #{e.message}"
        end
      }
    end
  end

  def reports
    @node = resource
    if params[:kind] == "inspect"
      @reports = @node.reports.inspections
      @tab = false
    elsif Node.possible_statuses.include?(params[:status])
      @reports = @node.reports.send(params[:status])
      @tab = params[:status]
    else
      @reports = @node.reports.applies
      @tab = 'all'
    end
    respond_to do |format|
      format.html { @reports = paginate_scope(@reports); render 'reports/index' }
    end
  end

  protected

  # Returns finder method for instantiating a resource using `params[:id]`
  def method_for_find
    # if the ID param is numeric treat it as a numeric ID
    # otherwise, treat it as the node name
    if params[:id] =~ /\A\d+\Z/
      :find
    else
      # throw a RecordNotFound exception when we can't find the node so Rails can return the correct 404 response code
      :find_by_name!
    end
  end

  # Render the index using the +scope_name+ (e.g. :successful for Node.successful).
  def scoped_index(*scope_names)
    index! do |format|
      scope = end_of_association_chain
      if params[:q]
        scope = scope.search(params[:q])
      end
      scope_names.each do |scope_name|
        scope = scope.send(scope_name)
      end
      set_collection_ivar(scope.with_last_report.by_report_date)

      format.html { render :index }
      format.json { render :json => collection }
      format.yaml { render :yaml => collection }
      format.csv {
        expires_in 0, 'must-revalidate' => true, :private => true
        render :csv => collection, :filename => "#{scope_names.join('-')}-nodes"
      }
    end
  end
end
